/* This file is a part of the program Kira.
 * Copyright (C) Johann Usovitsch <jusovitsch@googlemail.com>
 * Philipp Maierhoefer <particle@maierhoefer.net>
 * Peter Uwer <peter.uwer@physik.hu-berlin.de>
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version , or (at
 * your option) any later version as published by the Free Software 
 * Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
*/
#ifndef COEFF_VEC_H
#define COEFF_VEC_H

#include <iostream>
#include <string>
#include <vector>
#include <utility>

namespace pyred
{

using cvec = std::vector<uint64_t>;

class Coeff_vec
{
  friend Coeff_vec operator-(const Coeff_vec&);
  friend Coeff_vec operator+(const Coeff_vec&, const Coeff_vec&);
  friend Coeff_vec operator-(const Coeff_vec&, const Coeff_vec&);
  friend Coeff_vec operator*(const Coeff_vec&, const Coeff_vec&);
  friend Coeff_vec operator/(const Coeff_vec&, const Coeff_vec&);
  friend bool operator==(const Coeff_vec&, const Coeff_vec&);
  friend std::ostream& operator<<(std::ostream&, const Coeff_vec&);
private:
  cvec c;
  static uint ncoeffs; // number of coefficients in coefficient vector
  static cvec prime;
  static bool auto_vars;
  static std::string set_to_one;
  static std::vector<std::pair<std::string,cvec>> randvars;
public:
  static void init(const std::vector<std::string>& = {}, uint = 1,
                   const std::string& = "");
  Coeff_vec();
  Coeff_vec(const Coeff_vec&);
  Coeff_vec(const std::string&);
  template<typename T> // only for integral types
  Coeff_vec(T);
  Coeff_vec& operator+=(const Coeff_vec&);
  Coeff_vec& operator-=(const Coeff_vec&);
  Coeff_vec& operator*=(const Coeff_vec&);
  Coeff_vec& operator/=(const Coeff_vec&);
  Coeff_vec& operator^=(const Coeff_vec&);
};

Coeff_vec operator-(const Coeff_vec&);
Coeff_vec operator+(const Coeff_vec&, const Coeff_vec&);
Coeff_vec operator-(const Coeff_vec&, const Coeff_vec&);
Coeff_vec operator*(const Coeff_vec&, const Coeff_vec&);
Coeff_vec operator/(const Coeff_vec&, const Coeff_vec&);
bool operator==(const Coeff_vec&, const Coeff_vec&);
std::ostream& operator<<(std::ostream&, const Coeff_vec&);

template<typename T>
inline Coeff_vec::Coeff_vec(T a): c(cvec()) {
  if (a >= 0) {
    c.resize(ncoeffs, a);
  }
  else {
    auto cc = uint64_t(-a);
    c.reserve(ncoeffs);
    for (const auto& p: prime) {
      c.emplace_back(p-cc);
    }
  }
}

}

#endif
