/* This file is a part of the program Kira.
 * Copyright (C) Johann Usovitsch <jusovitsch@googlemail.com>
 * Philipp Maierhoefer <particle@maierhoefer.net>
 * Peter Uwer <peter.uwer@physik.hu-berlin.de>
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version , or (at
 * your option) any later version as published by the Free Software 
 * Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
*/
#ifndef DATABASE_H
#define DATABASE_H

#include <string>
#include <tuple>

#include "integral.h"
#include "sqlite3/sqlite3.h"

#define NDBINT 8


struct DBintegral{
  std::string indices;
  int topology;
  std::string coefficient;
  int massDimension;
  long unsigned id;
};

class DataBase {
public:
  DataBase(std::string name);
  ~DataBase();
  void execute_statement(std::string& sql,std::string message);
  int checkTable();
  void create_integral_table(int l_Indices);
  void create_equation_table();
  void begin_transaction();
  void commit_transaction();
  
  void prepare_pyred();
  void prepare_find_master();
  void bind_get_answer(int topology,int sector,int dots,int nums,std::vector<long unsigned>& mandatory);
  
  void prepare_integral();
  void bind_integral(SEEDIntegral& integral, int indices);
  
  void prepare_backsubstitution();
  void bind_equation(Integral& integral, int length, long unsigned ID);
  
  void prepare_id();
  long unsigned bind_get_id(std::vector<int> & indices, int topology);
  
  void prepare_lookup_id();
  int bind_id_get_integral(long unsigned id, std::tuple<std::string,unsigned,unsigned>& integral);
  void bind_id_get_equation(long unsigned id, std::vector<DBintegral>& equation);
  
  void finalize();
  
  int number_of_columns();
  std::vector<std::string> columnName;
private:
  sqlite3 *db;
  char *errorMessage;
  int rc;
  std::string sql;
  sqlite3_stmt* stmt;
  int l_Indices;
  unsigned int totalColumn;
};
#endif // DATABASE_H