/* This file is a part of the program Kira.
 * Copyright (C) Johann Usovitsch <jusovitsch@googlemail.com>
 * Philipp Maierhoefer <particle@maierhoefer.net>
 * Peter Uwer <peter.uwer@physik.hu-berlin.de>
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version , or (at
 * your option) any later version as published by the Free Software 
 * Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
*/

//jule is the number of propagators
#include "yaml-cpp/yaml.h"
#include "sort_rules.h"
#include <fstream>
#include "kira.h"
#include "integral.h"
#include "ReadYamlFiles.h"
#include "tools.h"
#include "sqlite3/sqlite3.h"
#include "dataBase.h"
#include <sys/stat.h>
#include "createSeeds.h"
#include "interface.h"
#include <omp.h>
#include <pthread.h>

#include <gzstream.h>

#include "integrals.h"

#include <algorithm>

using namespace pyred;
  
using namespace std;
using namespace GiNaC;
using namespace YAML;

static pthread_mutex_t mPreserveFermat;
static pthread_cond_t cPreserveFermat;

static Loginfo & logger = Loginfo::instance();

Kira::Kira (string name, int coreNumber_,int algebra_, string pyred_config_) {
  
  coreNumber = coreNumber_;
  algebra = algebra_;
  jobName = name;
  pyred_config = pyred_config_;
};

int Kira::look_up_symmetry(vector<Equation*>& setUpEq, IBPIntegral *&integral, SYM *&sym,int topologyNUM){
  
  ex coef1, coef2, number;
  IBPVG Identitie_Integral, Identitie_Integral2;
  int sec=integral->characteristics[SECTOR];
  
  for(vector<symmetries>::iterator symHandler = sym[sec].begin(); symHandler!=sym[sec].end(); symHandler++){

    if (symHandler->det == 0)
      continue;
    
    if (symHandler->symDOTS == 1 && (integral->characteristics[NUM]>0 /*|| (integral->characteristics[DOTS]-integral->characteristics[DENCOUNT])>1*/)){
      continue;
    }
    
//     if (symHandler->symDOTS == 0 && (integral->characteristics[DOTS]-integral->characteristics[DENCOUNT])>0)
//       continue;
    
//     cout<<symHandler->symDOTS<<" "<<(integral->characteristics[DOTS]-integral->characteristics[DENCOUNT])<< " " << integral->characteristics[NUM] << "\n";
    
    string topoName = collectReductions[ symHandler->topology ];
    string topoNameOriginal = collectReductions[topologyNUM];
    
    vector<vector<pair<ex,int> > > exTerm;
    ex ausd2;
    
    for (int j=0; j<topology[topoNameOriginal].jule;j++){
      if (symHandler->ing[j]<0){
        vector<pair<ex,int> > uj;
        if (integral->indices[j]<0){
          ausd2 = topology[topoNameOriginal].props[j].expand().subs(symHandler->subst,
                            subs_options::algebraic).expand();
          fs<lst>(ausd2,kinematic);
          ausd2 = ausd2.subs(topology[topoName].scal2Props[symHandler->externalSymmetry], subs_options::algebraic);
          ex traley;
          if((traley=ausd2.subs(suby,subs_options::algebraic))!=0){
            uj.push_back(pair<ex, int > (traley,-1));
          }
          for(int jopp=0; jopp<topology[topoNameOriginal].jule;jopp++){
            uj.push_back(pair<ex, int > (ausd2.coeff(topology[topoNameOriginal].propSymb[jopp],1),jopp));
          }
        }
        for (int iA=0; iA < -integral->indices[j];iA++){
          exTerm.push_back(uj);
        }
      }
    }
    
//     for (int j=0; j<topology[topoNameOriginal].jule;j++){
//       vector<pair<ex,int> > uj;
// //       if (integral->indices[j]<0){
//         ausd2 = topology[topoNameOriginal].props[j].expand().subs(symHandler->subst,
//                           subs_options::algebraic).expand();
//         fs<lst>(ausd2,kinematic);
//         ausd2 = ausd2.subs(topology[topoName].scal2Props[symHandler->externalSymmetry], subs_options::algebraic);
//         ex traley;
//         if((traley=ausd2.subs(suby,subs_options::algebraic))!=0){
//           uj.push_back(pair<ex, int > (traley,-1));
//         }
//         for(int jopp=0; jopp<topology[topoNameOriginal].jule;jopp++){
//           uj.push_back(pair<ex, int > (ausd2.coeff(topology[topoNameOriginal].propSymb[jopp],1),jopp));
//         }
// //       }
// //       for (int iA=0; iA < -integral->indices[j];iA++){
//         exTerm.push_back(uj);
// //       }
//     }
    
    
    
    IBPIntegral *newIntegral;  
    newIntegral = new IBPIntegral(topology[topoNameOriginal].jule);
    
    for (int j=0; j < topology[topoNameOriginal].jule;j++){
      newIntegral->indices[j] = 0;
    }
    for (int j=0; j<topology[topoNameOriginal].jule;j++){
      if (symHandler->ing[j] >= 0){
        newIntegral->indices[symHandler->ing[j]] = integral->indices[j];
      }
    }
    newIntegral->coefficient=1;
    
    newIntegral->characteristics[TOPOLOGY] = symHandler->topology;
    newIntegral->characteristics[SECTOR] = symHandler->sector;
    
    Identitie_Integral.push_back(newIntegral);
      
    for(vector<vector<pair<ex,int> > >::iterator kii=exTerm.begin();
        kii!=exTerm.end();kii++){
      for(ItIBPVG i = Identitie_Integral.begin(),
            end = Identitie_Integral.end(); i != end ; ++i){
        for(vector<pair<ex,int> >::iterator lii=(*kii).begin();
            lii!=(*kii).end();lii++){

          newIntegral=new IBPIntegral(topology[topoNameOriginal].jule);
          
          for ( int j = 0; j<topology[topoName].jule; j++) {
            newIntegral->indices[j] = (*i)->indices[j];
          }
          if((*lii).first!=0){
            if((*lii).second!=-1)
              --newIntegral->indices[(*lii).second];
            newIntegral->coefficient = 
              (((*lii).first)*(*i)->coefficient).expand();
            newIntegral->generate_characteristics(symHandler->topology);
            Identitie_Integral2.push_back(newIntegral);
          }
          else
            delete newIntegral;
        }
        delete (*i);
      }
      Identitie_Integral.clear();
      Identitie_Integral=Identitie_Integral2;
      Identitie_Integral2.clear();
    }
    newIntegral = new IBPIntegral(topology[topoNameOriginal].jule);
    for ( int j = 0; j<topology[topoNameOriginal].jule; j++) {
      newIntegral->indices[j] = integral->indices[j];
    }
    newIntegral->coefficient=-1;
    newIntegral->characteristics[TOPOLOGY] = topologyNUM;
    newIntegral->characteristics[SECTOR] = sec;
    Identitie_Integral.push_back(newIntegral);

    Equation * tokenEq = new Equation(Identitie_Integral,topology[topoNameOriginal].jule,(int_seeds->integralMap));
    
    if((tokenEq->sort())>1){
      setUpEq.push_back(tokenEq);
      numberOfEqXXX++;
    } 
    else {
      delete tokenEq;
    }
    
    for ( ItIBPVG i = Identitie_Integral.begin(); i != Identitie_Integral.end(); i++) {
      delete *i;
    }
    Identitie_Integral.clear();
  }
  return 1;
};

int Kira::setup_sym_eq(vector<Equation*>& setUpEq, string name, int topologyNUM, int secNUM){
  
  if(!(int_seeds->aIBPint[topologyNUM][secNUM].size())){
    return 0;
  }
  
  logger << "Symmetry seeds: "<< int_seeds->aIBPint[topologyNUM][secNUM].size() << ";";
  
  for  ( auto it = int_seeds->aIBPint[topologyNUM][secNUM].begin(); it != int_seeds->aIBPint[topologyNUM][secNUM].end(); ++it) {
    
    IBPIntegral *integral = new IBPIntegral((*it),integralfamily.jule);
    
    int sec=integral->characteristics[SECTOR];
    
    if(topology[collectReductions[topologyNUM]].symVec[sec].size() && (integral->characteristics[NUM]<3)){//<3
      look_up_symmetry(setUpEq, integral,topology[collectReductions[topologyNUM]].symVec,topologyNUM);
    }
    
    if( topology[collectReductions[topologyNUM]].relVec[sec].size() && (integral->characteristics[NUM]<3) ){ // <2
      look_up_symmetry(setUpEq, integral,topology[collectReductions[topologyNUM]].relVec,topologyNUM);
    }

    delete integral;        
  }
  
  return 1;
  
};

int Kira::setup_ibp_li_eq(vector<Equation*>& setUpEq, VE &identities, vector<long unsigned>& weights, string name, int topology){
  
  size_t nID=weights.size();
  if(!(identities.size())||!nID){
    return 0;
  }
  
  omp_set_dynamic(0);
  int nThreads = coreNumber;
  omp_set_num_threads(nThreads);
  
  omp_lock_t lck;
  omp_init_lock(&lck);
  
  
  #pragma omp parallel for schedule(dynamic) shared(lck)    
  for  (unsigned it = 0; it < nID; it++) {
    
    for(unsigned idenIt = 0; idenIt < identities.size(); idenIt++){
      
      int id = -1;
      pthread_mutex_lock(&mPreserveFermat);
        while(!core_mask)
          pthread_cond_wait(&cPreserveFermat, &mPreserveFermat);

        for(int i=0;i<coreNumber;i++){
          if(((core_mask>>i)&1LL)==1LL){
            id=i;
            core_mask &= ~(1LL<<i);
            break;
          }
        }
      pthread_mutex_unlock(&mPreserveFermat);
        
      Equation* tokenEq = new Equation(*identities[idenIt],weights[it],(int_seeds->integralMap),topology, fermat, id);
      
      pthread_mutex_lock(&mPreserveFermat);
        core_mask |= 1LL<<id;
      pthread_cond_broadcast(&cPreserveFermat);
      pthread_mutex_unlock(&mPreserveFermat);
      
      if(tokenEq->sort()){
      
        omp_set_lock(&lck);

          setUpEq.push_back(tokenEq);
          
          numberOfEqXXX++;
        
        omp_unset_lock(&lck);        
      }
      else{

        delete tokenEq;
      }
    }
  }

  omp_destroy_lock(&lck);
  return 1;
  
};

void Kira::initiate_system_of_equations(int nOfTopology){
  
  Clock clock;
  
  numberOfEq=0;
  
  ofstream output;
  mkdir((outputDir+"/tmp/"+integralfamily.name).c_str(), S_IRWXU | S_IRWXG | S_IROTH | S_IXOTH);
  output.open((outputDir+"/tmp/"+integralfamily.name+"/SYSTEMconfig").c_str());
  
  logger << "***********************************************************\n";
  logger << "Generate equations\n";
  logger << "***********************************************************\n";
  
  long unsigned lastID=0;
  
  for(int i = 0; i < nOfTopology; i++){
    
    for(int j = 0; j < int_seeds->nSector[i]; j++){
      
      vector<Equation*> setUpEq;
      
      int secNUM = int_seeds->mappingSectorReverse[i][j];
	
      if(!(topology[collectReductions[i]].identitiesIBP.size())||!(int_seeds->aIBPint[i][j]).size() ){
        continue;
      }
      
      logger << "Topology: " << collectReductions[i] << "; ";
      
      logger << "Sector: " << secNUM << "; ";
      
      numberOfEqXXX = 0;
      
      setup_sym_eq(setUpEq,"SYM",i,j); 
      
      logger <<"\n";
      
      sort(setUpEq.begin(), setUpEq.end(), sort_rules4);
      
      write_init_system(setUpEq, collectReductions[i],secNUM,"sym");
      
      if(setUpEq.size()) {
        lastID = (setUpEq.back())->equation[0].id;
      }
      
      for ( unsigned it = 0; it < setUpEq.size(); it++) {
	  delete setUpEq[it];
      }
      numberOfEq += static_cast<unsigned>(setUpEq.size());
      setUpEq.clear();
    }
  }
  
  for(int i = 0; i < nOfTopology; i++){
    
    omp_set_nested(1);
    omp_set_dynamic(0);
    int nThreads = coreNumber;
    if(maxParallelSectors)
      omp_set_num_threads(maxParallelSectors);
    else
      omp_set_num_threads(nThreads);
    
    #pragma omp parallel for schedule(dynamic)
    for(int j = 0; j < int_seeds->nSector[i]; j++){
      
      Clock clock2;
      
      vector<Equation*> setUpEq;
      
      int secNUM = int_seeds->mappingSectorReverse[i][j];
	
      if(!(topology[collectReductions[i]].identitiesIBP.size())||!(int_seeds->aIBPint[i][j]).size() ){
        continue;
      }
      
      #pragma omp critical
      {
        logger << "Topology: " << collectReductions[i] << "; ";
      
        logger << "Sector: " << secNUM << "; ";
        logger << " seeds: " << int_seeds->aIBPint[i][j].size() << "; \n";
      }
      
      setup_ibp_li_eq(setUpEq,topology[collectReductions[i]].identitiesIBP, int_seeds->aIBPint[i][j], "IBP", i);
      
      setup_ibp_li_eq(setUpEq,topology[collectReductions[i]].identitiesLI, int_seeds->aIBPint[i][j], "LI", i);
      
      sort(setUpEq.begin(), setUpEq.end(), sort_rules4);
      write_init_system(setUpEq,collectReductions[i],secNUM,"");
      
      #pragma omp critical
      {
        numberOfEq+=static_cast<unsigned>(setUpEq.size());
        logger << "Topology: " << collectReductions[i] << "; ";
        logger << "Sector: " << secNUM << "; ";
        logger << "( " << clock2.eval_time() << " s )\n";
        if(setUpEq.size() && setUpEq.back()->equation[0].id > lastID)
          lastID = (setUpEq.back())->equation[0].id;
      }
      
      for ( unsigned it = 0; it < setUpEq.size(); it++) {
	  delete setUpEq[it];
      }
      setUpEq.clear();
    }
  }

  output<<numberOfEq<<"\n";
  output<<lastID<<"\n";
  
  output.close();
  
  logger << "\n( " << clock.eval_time() << " s )\n";
};


vector<vector<int > > Kira::read_id2int(){
  
  ifstream fileInput;
  fileInput.open((outputDir+"/results/"+integralfamily.name+"/id2int").c_str());
  
  int jule = integralfamily.jule;
  string line;
  vector<vector<int> > arraySeed;
  
  while(getline(fileInput, line)) {
    
    //get string indices
    if(line.size()<7)
      continue;
    
    size_t found;
    
    if((found = line.find_first_of("#")) != string::npos){
      logger << "this line will be skept: " << line<< "\n";
      continue;
    }
    
    if((found = line.find_first_of("[")) == string::npos){
      logger << "this line will be skept: " << line<< "\n";
      continue;
    }
    
    line.erase(line.begin(),line.begin()+found+1);
    line.erase(remove_if(line.begin(), line.end(), ::isspace), line.end());
    
    if((found = line.find_first_of("]")) == string::npos){
      logger << "this line will be skept: " << line<< "\n";
      continue;
    }
    
    line.erase(line.begin()+found);//indices
    
    //convert string indices to integer array
    int indicesCounter;
    vector<int> inputSeed;
    int countN=0;
    while((found = line.find_first_of(",")) != string::npos){

      if((jule + 1)!=countN && (jule + 2)!=countN){
	countN++;
	line = line.substr(found+1);
	continue;
      }
      
      istringstream(line.substr(0,found))>>indicesCounter;
      inputSeed.push_back(indicesCounter);
      line=line.substr(found+1);
      
      countN++;
    }
    
    istringstream(line) >> indicesCounter;
    inputSeed.push_back(indicesCounter);
    arraySeed.push_back(inputSeed);
  }
  
  return arraySeed;
}

void Kira::run_pyred(std::vector<long unsigned>& mandatory, std::vector<long unsigned>& optional,
  vector<info> mandatoryRec, vector<info> optionalRec){
  
  Clock clock;
  logger << "\n***** Starting PYRED **************************************\n";
  
  
  pyred::Config::parse(pyred_config);
  pyred::Config::set_verbosity(2);
  
  vector<vector<pair<pyred::intid,string> > > eqs;
  vector<vector<tupelCharacteristics > >  characteristics;
  vector<pyred::intid> independent_eqnums_OTF;
  
  if(mandatoryRec.size()==0 && mandatory.size() == 0){
    
    logger << "\n***** Do not select equations *****************************\n";
    
    if(OTF){
      load_pyred_on_the_fly(independent_eqnums_OTF,mandatory,optional);
    }
    else {
//       auto arraySeed = read_id2int();
      
      load_pyred(eqs);

      auto sys = pyred::System(move(eqs));

      auto independent_eqnums = sys.solve();
      
      if( writeNumericalSystem )
        get_reduced_eqs(sys);
      auto eqs = sys.retrieve(independent_eqnums);
      
      write_pyred(eqs/*,arraySeed*/,independent_eqnums);
      
    }
  }
  else{
    
    logger << "\n***** Select equations recursively ************************\n";
    DataBase* database = new DataBase(outputDir+"/results/"+collectReductions[integralfamily.topology]+"/kira.db");
    if(database[0].number_of_columns()!=0){
      
      database[0].prepare_pyred();
      
      for(size_t it3 = 0; it3 < mandatoryRec.size(); it3++){
      
	database[0].bind_get_answer(mandatoryRec[it3].topology,mandatoryRec[it3].sector,
					mandatoryRec[it3].dots,mandatoryRec[it3].nums,mandatory);
      }
      for(size_t it3 = 0; it3 < optionalRec.size(); it3++){

	database[0].bind_get_answer(optionalRec[it3].topology,optionalRec[it3].sector,
					optionalRec[it3].dots,optionalRec[it3].nums,optional);
      }      
    }
    database[0].finalize();
    delete database;
    
    sort( mandatory.begin(), mandatory.end() );
    mandatory.erase( unique( mandatory.begin(), mandatory.end() ), mandatory.end() );
    sort( optional.begin(), optional.end() );
    optional.erase( unique( optional.begin(), optional.end() ), optional.end() );
    
//     for(unsigned it =0; it< mandatory.size(); it++)
//       cout << mandatory[it] << " ";
    
    logger << "length of mandatory list: " << mandatory.size() << "\n";
    logger << "length of optional list: " << optional.size() << "\n";
    
    if(OTF){
      load_pyred_on_the_fly(independent_eqnums_OTF,mandatory,optional);
    }
    else{
      
//       auto arraySeed = read_id2int();
      
      load_pyred(eqs);

      auto sys = pyred::System(move(eqs));

      auto independent_eqnums = sys.solve();

      auto subsys = sys.select(mandatory,optional);
      
      logger << "Reduce: " << subsys.first.size() << " instead of: "
	      << independent_eqnums.size() << " integrals\n";
      
      record_masters(subsys);
      
      if( writeNumericalSystem )
        get_reduced_eqs(sys);
      
      auto eqs = sys.retrieve(subsys.first);
      
      write_pyred(eqs/*,arraySeed*/,subsys.first);
    }
  }
  logger << "( " << clock.eval_time() << " s )\n";
};

void Kira::complete_triangular(){
  
  logger << "\n***** Bring system in triangular form *********************\n";

  load_triangular();
  
  run_reduction(0);
  write_triangular();
  
  // Clean Up
  for( FWREDI toIt = forwardRed.begin(); toIt != forwardRed.end(); toIt++){ //collect all keys to iterate
    
    if(get<1>(toIt->second)>0){

      delete [] get<0>(toIt->second)[0][0];
    }
    
    delete [] get<0>(toIt->second)[0];
    delete [] get<0>(toIt->second);
  }
  forwardRed.clear();
  rdy2F.clear();
  reduce_please.clear();
  
  
};

void Kira::complete_reduction(){
  
  logger << "\n***** Run the back substitution ***********************\n";
  
  string inputName = outputDir+"/tmp/"+integralfamily.name+"/masters";
  ifstream input;
  
  vector<unsigned long> masterVector;
  
  if(file_exists(inputName.c_str())){  
    input.open(inputName.c_str());
    while(1){
      long unsigned ID;
      if(!(input >>ID)) break;
      masterVector.push_back(ID);
    }
    input.close();
  }
  
  auto ret = selectMastersReduction.equal_range(integralfamily.name);
  
  if (ret.first != selectMastersReduction.end()){
  
    for (auto itRet=ret.first; itRet!=ret.second; ++itRet){
          
      for(unsigned i = 0; i < masterVector.size(); i++){
          
        auto findMaster = find( (itRet->second).begin(),(itRet->second).end(), i+1 );
        
        if( findMaster == (itRet->second).end() ){
          
          masterVectorSkip.push_back( masterVector[i] );
        }
      }
      
      load_back_substitution();
      back_subs();
      clean_back_subs();
    }
  }
  else{
    load_back_substitution();
    back_subs();
    clean_back_subs();
  }
};

void Kira::clean_back_subs(){
  
  for ( unsigned it = 0, end = reihen+1; it != end; it++) {
    if (length[it]) {
      delete [] allEq[it][0];
    }
    if(allEq[it]!=0)
      delete [] allEq[it];
  }
  
  delete [] rdy2P;
  delete [] last_reduce;
  delete [] length;
  delete [] allEq;
  reduct2StartHere.clear();
  occurrence.clear();
  reverseLastReduce.clear();
  masterVectorSkip.clear();
}


void Kira::print_equation2(Integral*& integral,ogzstream& output){
  
  output << "Eq" << "\n";
  output << integral[0].id << "\n";
  output << integral[0].length << "\n";
  
  for (unsigned j=0; j<integral[0].length;j++){
    integral[j].length=integral[0].length;
    output << integral[j];
  }
}


void Kira::print_equation(Integral*& integral,ogzstream& output,pair<int,int>& printControl,const string & dest){
  
  if (integral[0].characteristics[SECTOR] != printControl.first 
    ||integral[0].characteristics[TOPOLOGY] != printControl.second){
    printControl.first = integral[0].characteristics[SECTOR];
    printControl.second = integral[0].characteristics[TOPOLOGY];
    output.close();
    string name = outputDir+"/tmp/" + dest + "_" + collectReductions[printControl.second] + "_" + something_string(printControl.first)+".gz";
    output.open (name.c_str());
  }
  
  output << "Eq" << "\n";
  output << integral[0].id << "\n";
  output << integral[0].length << "\n";
  
  for (unsigned j=0; j<integral[0].length;j++){
    integral[j].length=integral[0].length;
    output << integral[j];
  }
}

void Kira::write_init_system(std::vector<Equation*>& setUpEq, const std::string & topology, int secNUM, const std::string & add){
  
  ogzstream output;
  
  output.open ((outputDir+"/tmp/"+integralfamily.name+"/SYSTEM"+add+"_"+topology+"_"+to_string(secNUM)+".gz").c_str());
  
  for( size_t it = 0; it < setUpEq.size(); it++){
    print_equation2(setUpEq[it]->equation,output);
  }
  
  output.close();
  
};

void Kira::write_pyred(vector<vector<pair<long unsigned,string> > > & eqs,/*
  vector<vector<int> > & arraySeed,*/ vector<pyred::intid> & independent_eqnums){
  
  logger << "\nWrite with PYRED reduced system to hard disk \n";
  logger << "(--> "<<outputDir<<"/tmp/NUMconfig, "<<outputDir<<"/tmp/NUM###) \n";
  
  ofstream output;
  
  output.open ((outputDir+"/tmp/"+integralfamily.name+"/NUMconfig").c_str());
  output << eqs[eqs.size()-1][0].first << "\n";
  
  pair<unsigned,unsigned> printControl;
  printControl.first = -1;
  printControl.second = -1;
  
  for(unsigned it = 0; it < eqs.size(); it++){
    long unsigned id = eqs[it][0].first;
    
    auto iglback = pyred::IntegralPyred(id);
    auto property = iglback.properties(id);
    
    if (property.sector != printControl.first //sector
    || property.topology != printControl.second){ //topology
      printControl.first = property.sector; //sector
      printControl.second = property.topology; //topology
      output.close();
      string name = outputDir+"/tmp/" + integralfamily.name + "/NUM" + "_" + collectReductions[printControl.second] + "_" + something_string(printControl.first);
      output.open (name.c_str());
    }
    output << "Eq" << "\n";
    output << eqs[it][0].first << "\n";
    output << eqs[it].size() << "\n";
  
    for (size_t it2 = 0; it2 < eqs[it].size(); it2++){
      output << eqs[it].size() << " ";
      output << eqs[it][it2].second << " ";
      output << eqs[it][it2].first << " ";
      output << property.sector << " ";
      output << property.topology << " ";
      output << "0" << "\n"; //flag2
    }
  }
  
  output.close();
};

void Kira::write_triangular(){
  
  logger << "\nWrite the triangular system to "<<outputDir<<"/tmp/ on hard disk\n";
  logger << "(--> "<<outputDir<<"/tmp/"<<integralfamily.name<<"/VERconfig, ";
  logger <<outputDir<<"/tmp/"<<integralfamily.name<<"/VER###) \n";
  
  ogzstream output;
  
  pair<int,int> printControl;
  printControl.first = -1;
  printControl.second = -1;
  
  int rememberID=0;
  for( FWREDI toIt = forwardRed.begin(); toIt != forwardRed.end(); toIt++){ //collect all keys to iterate
    if(get<1>(toIt->second)>0 && get<0>(toIt->second)[0][0][0].flag2 == 0){
      print_equation(get<0>(toIt->second)[0][0],output,printControl,integralfamily.name+"/VER");
      rememberID++;
    }
  }
  output.close();
  
  ofstream outputConfig;
  
  outputConfig.open ((outputDir+"/tmp/"+integralfamily.name+"/VERconfig").c_str());
  outputConfig << rememberID << "\n";
  outputConfig.close();
};


void Kira::read_integral2(igzstream& input, int eqLength){
    
  Integral *integral = new Integral [eqLength];
  
  for (int j=0; j<eqLength;j++){
    
    if(!(input>>integral[j])) break;

    integral[j].length=eqLength;
  }
  FWREDI mapContent = forwardRed.find(integral[0].id);
  if ( mapContent != forwardRed.end() ) {
    
    get<0>(mapContent->second)[0][get<1>(mapContent->second)]=integral;
    get<1>(mapContent->second)++;    
  }
  else{
    Integral *** put2map = new Integral ** [1];
    put2map[0] = new Integral * [RED];
    put2map[0][0]=integral;
    forwardRed.insert(pair<long unsigned, tuple<Integral***, int> > (integral[0].id, make_tuple(put2map,1)));
    rdy2F.insert(pair<long unsigned, int> (integral[0].id, 0));
  }
}

void Kira::read_integral(Integral*& integral, igzstream& input,int eqLength){
    
  integral = new Integral [eqLength];
  
  for (int j=0; j<eqLength;j++){
    
    if(!(input>>integral[j])) break;

    integral[j].length=eqLength;
    
    auto findSkipMaster = find(masterVectorSkip.begin(), masterVectorSkip.end(), integral[j].id);
    if ( findSkipMaster != masterVectorSkip.end() ) {
      integral[j].coefficientString = "0";
    }
    
    auto occContent = occurrence.find(integral[j].id);
    if ( occContent != occurrence.end() ) {
      (occContent->second)++;
    }
    else{
      occurrence.insert(pair<long unsigned, unsigned > (integral[j].id, 1) );
    }
  }
}


void Kira::load_pyred_on_the_fly(vector<long unsigned> & independent_eqnums,
		std::vector<long unsigned>& mandatory, std::vector<long unsigned>& optional){
  
  logger << "Load all files from "<<outputDir<<"/tmp/"<<integralfamily.name<<"\n";
  
  ifstream inputConfig((outputDir+"/tmp/"+integralfamily.name+"/SYSTEMconfig").c_str());
  
  inputConfig >> numberOfEq;
  
  inputConfig.close();
  
  if(skipPyred){
    ofstream output;
    output.open ((outputDir+"/tmp/"+integralfamily.name+"/independentEQS").c_str());
    for(unsigned i = 0; i < numberOfEq; i++) {
      output << i << "\n";
    }
    output.close();
  }
  else{
  
  
    auto sys = pyred::System();
    sys.reserve(numberOfEq);
    
    int eqLength; 
    long unsigned ID;
    string line;
    vector<string> add;
    add.push_back("sym");
    add.push_back("");
    
    for(unsigned k = 0; k < collectReductions.size(); k++) {
      
      for (int ii = 0; ii < (1<<integralfamily.jule)+1; ii++) {
        int sectorNumber = 0;
        for (int j = 0; j<integralfamily.jule; j++) {
          if ((1<<j) & ii)
            sectorNumber += (1<<j);
        }
        for(int ll = 0; ll<2; ll++){
          
          string inputName = outputDir+"/tmp/" + integralfamily.name + "/SYSTEM"+add[ll]+"_"+collectReductions[k] + "_" + something_string(sectorNumber)+ ".gz";
          
          if(file_exists(inputName.c_str())){  
            igzstream input;
            input.open(inputName.c_str());
            while(1){
              
              vector< pair<long unsigned, string > > eq1;
              
              if(!(input >>line)) break;
              if(!(input >>ID)) break;
              if(!(input >> eqLength)) break;
              
              for (int j=0; j<eqLength;j++){
                
                int skip;
                string coef;
                long unsigned id;
                
                if(!(input>>skip)) break;
                if(!(input>>coef)) break;
                if(!(input>>id)) break;
                
                eq1.push_back(make_pair(id, coef));
                
                if(!(input>>skip)) break; //sector
                if(!(input>>skip)) break; //topology
                if(!(input>>skip)) break; //flag2
              }
              sys.add_forward(eq1);
            }
            input.close();
          }
          load_bar(ii+1,(1<<integralfamily.jule)+1,50,100);
        }
      }
      
      logger << "\n";
    }
    independent_eqnums = sys.backward();
    
    
    logger << "\nWrite with PYRED_OTF reduced system to hard disk \n";
    logger << "(--> "<<outputDir<<"/tmp/"<<integralfamily.name<<"/independentEQS)\n";
    
    if(mandatory.size()>0){
      
      auto subsys = sys.select(mandatory,optional);
      
      logger << "Reduce: " << subsys.first.size() << " instead of: "
              << independent_eqnums.size() << " integrals\n";
              
      
      record_masters(subsys);
      
      ofstream output((outputDir+"/tmp/"+integralfamily.name+"/independentEQS").c_str());
      
      for(unsigned i = 0; i < subsys.first.size(); i++) {
      
        output << subsys.first[i] << "\n";
      }
      
      output.close();
    }
    else{

      ofstream output((outputDir+"/tmp/"+integralfamily.name+"/independentEQS").c_str());
      
      for(unsigned i = 0; i < independent_eqnums.size(); i++) {
      
        output << independent_eqnums[i] << "\n";
      }
      
      output.close();
    }
    
    if( writeNumericalSystem )
      get_reduced_eqs(sys);
  }  
};

void Kira::get_reduced_eqs(pyred::System & sys){
  
  auto& content = sys.get_content();
  
  ofstream output((outputDir+"/results/"+integralfamily.name+"/reducedEqs").c_str());
  
  for (auto itContent = content.begin(); itContent != content.end(); ++itContent) {
    
    output << "integral ID: " << ((*itContent).first) << " " <<"\n";
    output << "masters ID: ";
    
    auto vecTupel = (std::get<1>((*itContent).second));
    
    for (auto iter = vecTupel.begin(); iter != vecTupel.end(); ++iter) {
      output << *iter << " ";
    }
    output << "\n";
  }
  
  output.close();
};

void Kira::record_masters(std::pair<std::vector<pyred::intid>,std::vector<long unsigned> >& subsys){
  
  ofstream output((outputDir+"/results/"+integralfamily.name+"/masters").c_str());
  ofstream tmpoutput((outputDir+"/tmp/"+integralfamily.name+"/masters").c_str());
      
  logger << "Master integrals: (" << subsys.second.size() << ")\n";
  output << "Master integrals: (" << subsys.second.size() << ")\n";
  
  for (unsigned i=0; i<subsys.second.size(); i++){

    tuple<string,unsigned,unsigned> integral;

    get_properties(subsys.second[i],integral);

    logger << subsys.second[i] << ": ";
    logger << collectReductions[get<1>(integral)] << ": ";
    logger << get<0>(integral)<< ": ";
    logger << get<2>(integral);
    logger << "\n";

    output << subsys.second[i] << ": ";
    output << collectReductions[get<1>(integral)] << ": ";
    output << get<0>(integral)<< ": ";
    output << get<2>(integral);
    output << "\n";
    
    tmpoutput << subsys.second[i] << "\n";
  }
  output.close();
}


void Kira::load_pyred(std::vector<std::vector<std::pair<long unsigned,std::string> > > & eqs){
  
  logger << "Load all files from "<<outputDir<<"/tmp/\n";
  ifstream input((outputDir+"/tmp/"+integralfamily.name+"/SYSTEMconfig").c_str());
  input >> numberOfEq;
  input.close();
  
  for(unsigned k = 0; k < collectReductions.size(); k++) {
    
    for (int ii = 0; ii < (1<<integralfamily.jule)+1; ii++){
      int sectorNumber = 0;
      for (int j = 0; j<integralfamily.jule; j++) {
	if ((1<<j) & ii)
	  sectorNumber += (1<<j);
      }
      string inputName = outputDir+"/tmp/"+integralfamily.name+"/SYSTEM" + "_"+collectReductions[k] + "_"+something_string(sectorNumber);
      if(file_exists(inputName.c_str())){  
	input.open(inputName.c_str());
	while(1){
	  int eqLength;
	  long unsigned ID;
	  string line;
	  if(!(input >>line)) break;
	  if(!(input >>ID)) break;
	  if(!(input >> eqLength)) break;
	  
	  vector< pair<long unsigned, string > > eq1;
	  
	  for (int j=0; j<eqLength;j++){
	    
	    int skip;
	    string coef;
	    long unsigned id;
	    
	    if(!(input>>skip)) break;
	    if(!(input>>coef)) break;
	    if(!(input>>id)) break;
	    
	    eq1.push_back(make_pair(id, coef));
	    
	    if(!(input>>skip)) break; //sector
	    if(!(input>>skip)) break; //topology
	    if(!(input>>skip)) break; //flag2
	  }
	  eqs.push_back(eq1);
	}
	input.close();
      }
      load_bar(ii+1,(1<<integralfamily.jule)+1,50,100);
    }
    
    logger << "\n";
  }
};


int Kira::load_triangular(){

  vector<unsigned> independent_eqnums;
  
  ifstream inputConfig;
  string inputNameOTF = (outputDir+"/tmp/"+integralfamily.name+"/independentEQS").c_str();
  if(file_exists(inputNameOTF.c_str())){
    inputConfig.open(inputNameOTF.c_str());
    logger << "read from: " << inputNameOTF << "\n";
    
    while(1){
      
      unsigned numberIND;
      
      if(!(inputConfig >> numberIND)) break;
      
      independent_eqnums.push_back(numberIND);
    }
    inputConfig.close();
  }
  
  if(independent_eqnums.size()==0)
    return 0;
  
  unsigned countEQ = 0;
  unsigned countALL = 0;
  Integral *integral = new Integral [1];
  
  vector<string> add;
  add.push_back("sym");
  add.push_back("");
  
  for(unsigned k = 0; k < collectReductions.size(); k++) {
    
    for (int ii = 0; ii < (1<<integralfamily.jule)+1; ii++){
      
      int sectorNumber = 0;
       
      for (int j = 0; j<integralfamily.jule; j++) {
	if ((1<<j) & ii)
	  sectorNumber += (1<<j);
      }
      for(int ll = 0; ll<2; ll++){
        string inputName = outputDir+"/tmp/"+integralfamily.name + "/SYSTEM"+add[ll] + "_"+collectReductions[k] + "_"+something_string(sectorNumber)+".gz";
        if(file_exists(inputName.c_str())){
          
          igzstream input;
          input.open(inputName.c_str());
          while(1){
            
            int eqLength;
            long unsigned ID;
            string line;
            if(!(input >> line)) break;
            if(!(input >> ID)) break;
            if(!(input >> eqLength)) break;
            
            if(countALL!=independent_eqnums[countEQ]){
              
              countALL++;
              
              for (int j=0; j<eqLength;j++){

                if(!(input>>integral[0])) break;
              }
              continue;
            }
            else{
              countALL++;
              countEQ++;
            }
            read_integral2(input, eqLength);
          }
          input.close();
        }
        load_bar(ii+1,(1<<integralfamily.jule)+1,50,100);
      }
    }
  }
  
  for( FWREDI toIt = forwardRed.begin(); toIt != forwardRed.end(); toIt++){ //collect all keys to iterate
    if(get<1>(toIt->second)>1){
      reduce_please.push_back(get<0>(toIt->second)[0][0]->id);
    }
  }
  delete [] integral;
  
  return 1;
};


void Kira::load_back_substitution(){
  
  logger << "Load all files from "<<outputDir<<"/tmp/\n";
  
  reduct2StartHere.clear();
  occurrence.clear();
  reverseLastReduce.clear();

  
  ifstream inputConfig((outputDir+"/tmp/"+integralfamily.name+"/VERconfig").c_str());
  
  inputConfig >> reihen;
  
  if(reihen == 0){
    logger << "The system you try to reduce contains no equations.\n";
    exit(1);
  }
  inputConfig.close();
    
  allEq = new Integral** [reihen+1];
  last_reduce = new long unsigned [reihen+1];
  length = new unsigned [reihen+1];
  rdy2P = new unsigned [reihen+1];
  
  for (unsigned i = 0; i < reihen+1; i++) {
    allEq[i] = new Integral* [1];
    length[i]=0;
    rdy2P[i]=0;
  }
  
  eqnum=0;
  vector <pair<unsigned,unsigned> > setofS;
  for(unsigned k = 0; k < collectReductions.size(); k++) {
    
    for (int ii = 0; ii <(1<<integralfamily.jule)+1; ii++) {
      unsigned sectorNumber = 0;
      for (int j = 0; j<integralfamily.jule; j++) {
	if ((1<<j) & ii){
	  sectorNumber += (1<<j);
        }
      }

      string inputName = outputDir+"/tmp/"+integralfamily.name+"/VER" + "_"+collectReductions[k] + "_"+something_string(sectorNumber)+".gz";
      
      if(file_exists(inputName.c_str())){  

	int num_ones=0;
	unsigned testSector=sectorNumber;

	for(size_t i = 0; i < SEEDSIZE; ++i, testSector >>= 1) {
	  if ((testSector & 1) == 1)
	    ++num_ones;
	}
	int it=0;
        igzstream input;
	input.open(inputName.c_str());
	while(1){
	  unsigned eqLength; 
          unsigned long ID;
	  string line;
	  if(!(input >>line)) break;
	  if(!(input >> ID)) break;
	  if(!(input >> eqLength)) break;
	  
	  read_integral(allEq[eqnum][0],input,eqLength);
	  
	  last_reduce[eqnum]=ID;
          reverseLastReduce.insert( pair<long unsigned, unsigned> (ID,eqnum) );
	  length[eqnum]++;
	  rdy2P[eqnum]=3;
	  
	  if(num_ones < SEEDSIZE && !(it++)){
	    int flagy = 0;
	    for(size_t itt=0;itt<setofS.size();itt++){
	      if((setofS[itt].first & sectorNumber) == setofS[itt].first && setofS[itt].second == k ){
		flagy = 1;
		break;
	      }
	    }
	    if(!flagy){
	      setofS.push_back(make_pair(sectorNumber,k));
	      reduct2StartHere.push_back(eqnum);
	    }
	  }
	  eqnum++;
	}
	input.close();
      }
    }
  }  
}

ConvertResult::ConvertResult(Kira& kira, string topologyName_, int topologyNumber_, string mathName, const string &inputDir_, vector<long unsigned> & idOfSeed){
  
  topologyName = topologyName_;
  topologyNumber = topologyNumber_;
  inputDir= inputDir_;
  
  string trivialSector = inputDir+"/sectormappings/"+topologyName+"/trivialsector";
  
  if(!(file_exists(trivialSector.c_str())) ){
    logger << "\n"<<inputDir<<"/sectormappings/"+topologyName+"/trivialsector is missing.\n";
    logger << "\nThis topology "<<topologyName<<" is not reduced yet.\n";
    logger << "\nUnable to export results to Mathematica\n";
    exit(-1);
  }
  
  trivialSectorInput.open(trivialSector.c_str());
  read_trivial_sector(); 
  
  if( !(file_exists(mathName.c_str())) ){
    logger << "\nMissing file with a list of integrals: " << mathName<<"\n";
    logger << "\nUnable to export results to Mathematica\n";
    exit(-1);
  }
  
  seedsInput.open( mathName.c_str() );
  
  look_up_seeds();
  
  database = new DataBase(inputDir+"/results/"+topologyName+"/kira.db"); 
  
  int totalColumn = database[0].number_of_columns();
 
  if(totalColumn)
    get_integral_id(idOfSeed);
}

ConvertResult::~ConvertResult(){
  
 delete database;
 
}

void ConvertResult::read_trivial_sector(){
  
  string line;
  getline(trivialSectorInput, line);
  size_t found;
  int sectorINT;
  while((found = line.find_first_of(",")) != string::npos){
    istringstream(line.substr(0,found)) >> sectorINT;
    trivSV.push_back(sectorINT);
    line=line.substr(found+1);
  }
  istringstream (line) >> sectorINT;
  trivSV.push_back(sectorINT);
  
};

int ConvertResult::get_sector_number(vector<int>& seed){
  int sectorNumber=0;
  int count=0;
  for(unsigned int iterator=0; iterator<seed.size();iterator++){
    if(seed[iterator]>0)
      sectorNumber+=(1<<count);
    count++;
  }
  return sectorNumber;
}


int ConvertResult::skip_integral(vector<int>& seed){
  
  int sectorNumber=get_sector_number(seed);

  vector<int>::iterator itSKIP = find (trivSV.begin(), trivSV.end(), sectorNumber);
  if ( itSKIP != trivSV.end() ) {
    
    logger.set_level(2);
    logger << "This integral is zero: \n";
    for(unsigned int iterator=0; iterator<seed.size();iterator++){
      logger << seed[iterator] << " ";
    }
    logger << "\n";
    logger.set_level(1);
    zeroIntegrals.push_back(seed);
 
    return 1;
    
  }
  return 0;

}

void ConvertResult::get_integral_id(vector<pyred::intid> &idOfSeed){  
  
  database[0].prepare_id();
  
  for(size_t it = 0; it!=arraySeed.size();it++){
    
    if(skip_integral(arraySeed[it])==1)
      continue;
    
    long unsigned id = database[0].bind_get_id(arraySeed[it],topologyNumber);
    
    if(id==0)
      continue;
    
    idOfSeed.push_back(id);
    
    logger.set_level(2);
    logger<< "This integral will be extracted: \n";
    
    for(unsigned int iterator=0; iterator<arraySeed[it].size();iterator++)
      logger << arraySeed[it][iterator] << " ";
    
    logger << "\n";
    logger.set_level(1);
    
  }
  database[0].finalize();
  
}

void ConvertResult::look_up_seeds(){

  logger << "Look up the seeds.\n";

  string line;
  
  logger.set_level(2);
      
  while(getline(seedsInput, line)) {
    
    //get string indices
    if(line.size()<4)
      continue;
    
    size_t found;
    
    if((found = line.find_first_of("#")) != string::npos){
      logger << "this line will be skept: " << line<< "\n";
      continue;
    }
    
    if((found = line.find_first_of("[")) == string::npos){
      logger << "this line will be skept: " << line<< "\n";
      continue;
    }
    
    line.erase(line.begin(),line.begin()+found+1);
    line.erase(remove_if(line.begin(), line.end(), ::isspace), line.end());
    
    if((found = line.find_first_of("]")) == string::npos){
      logger << "this line will be skept: " << line<< "\n";
      continue;
    }
    
    line.erase(line.begin()+found);//indices
    
    //convert string indices to integer array
    int indicesCounter;
    vector<int> inputSeed;
    
    while((found = line.find_first_of(",")) != string::npos){
      
      istringstream(line.substr(0,found))>>indicesCounter;
      inputSeed.push_back(indicesCounter);
      line=line.substr(found+1);
      
    }
    
    istringstream(line) >> indicesCounter;
    inputSeed.push_back(indicesCounter);
    arraySeed.push_back(inputSeed);
    
  }
  logger.set_level(1);
}

void ConvertResult::reconstruct_mass( Kira& kira, vector<DBintegral>& integralV ){

  for(size_t iterator=0;iterator!=integralV.size();iterator++){
    
    int massDimension = integralV[iterator].massDimension-integralV[0].massDimension;

    Perl2Kira perlKira(kira.invar,kira.massSet2One,kira.invarDim,kira.massSet2OneDim);
    perlKira.pipe_kira();
    perlKira.put_pipe(integralV[iterator].coefficient); 
    perlKira.close_pipe();
    perlKira.read_pipe(integralV[iterator].coefficient);
    stringstream ss;
    ss <<"*"<< kira.massSet2One<<"^("<<massDimension*2<<"/"<<kira.massSet2OneDim<<")";
    string MD = ss.str();
    integralV[iterator].coefficient="("+integralV[iterator].coefficient+")"+MD;
    kira.fermat[0].fermat_collect(const_cast<char*>(integralV[iterator].coefficient.c_str()));
    kira.fermat[0].fermat_calc();
    integralV[iterator].coefficient.assign(kira.fermat[0].g_baseout);
  }
}

void ConvertResult::prepare_FORM( vector<DBintegral>& integralV ){
  
  for(size_t iterator=0;iterator!=integralV.size();iterator++){
    
    Perl2Kira perlKira;
    perlKira.pipe_kira();
    perlKira.put_pipe(integralV[iterator].coefficient); 
    perlKira.close_pipe();
    perlKira.read_pipe(integralV[iterator].coefficient);
  }
}

int ConvertResult::output(Kira& kira, int massReconstruction, vector<long unsigned>& idOfSeed, int choice){  
  
  kiraOutput outString(choice);
  
  string reducefile = inputDir+"/results/" +topologyName + "/kira.db";
  
  if( !(file_exists(reducefile.c_str())) ){
    logger << "\nMissing file " << reducefile << ".\n";
    logger << "\nUnable to export results to "<<outString.str[0]<< "\n";
    return 0;
  }
  
  if( !database[0].checkTable()){
    logger << "\nNo reductions found in " << reducefile << ".\n";
    logger << "\nUnable to export results to "<<outString.str[0]<< "\n";
    return 0;
  }
    
  
  string mathematicafile = inputDir+"/results/" +topologyName + "/kira" + outString.str[1];
  Output.open( mathematicafile.c_str() );

  Output << outString.str[2];
    
  for(unsigned int it=0; it!=zeroIntegrals.size(); it++){
    
    Output << outString.str[3] << kira.collectReductions[topologyNumber]<<outString.str[5];
    
    for(unsigned int iterator=0; iterator<zeroIntegrals[it].size();iterator++){
    
      Output <<zeroIntegrals[it][iterator];
     
      if(iterator!=zeroIntegrals[it].size()-1)
	Output<< ",";
    }
    Output << outString.str[7]<<outString.str[4]<<outString.str[10];
    if(it!=zeroIntegrals.size()-1)
      Output << outString.str[6];
  }
  
  database[0].prepare_lookup_id();
  
  int counter=0;
  for(size_t it=0; it!=idOfSeed.size(); it++){
    vector<DBintegral> integralV;
    database[0].bind_id_get_equation(idOfSeed[it],integralV);
    
    if(integralV.size()>0)
      counter++;
    
    if(counter==1 && zeroIntegrals.size()>0 && integralV.size()>0)
      Output <<outString.str[6];
    
    if(counter>1 && integralV.size()>0)
      Output <<outString.str[6];
    
    if(massReconstruction)
      reconstruct_mass( kira, integralV );
    
    if(choice == 0)
      prepare_FORM(integralV);
    
    for(size_t iterator=0;iterator!=integralV.size();iterator++){
    
      if(iterator == 0)
	Output << outString.str[3] << kira.collectReductions[integralV[iterator].topology] <<outString.str[5];
      else
	Output << outString.str[4]<<kira.collectReductions[integralV[iterator].topology]<<outString.str[5];
      
      Output <<integralV[iterator].indices;
      
      if(iterator==0)
	Output <<outString.str[7];
      else
	Output <<outString.str[8]<<integralV[iterator].coefficient<<outString.str[9];
    }
    
    if(integralV.size()==1)
      Output <<outString.str[10];
    
  }
  
  database[0].finalize();
  
  Output << outString.str[11]<< "\n";
  logger << "\n";
  return 1;
};