/* This file is a part of the program Kira.
 * Copyright (C) Johann Usovitsch <jusovitsch@googlemail.com>
 * Philipp Maierhoefer <particle@maierhoefer.net>
 * Peter Uwer <peter.uwer@physik.hu-berlin.de>
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version , or (at
 * your option) any later version as published by the Free Software 
 * Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
*/
#include "kira.h"
#include "connect2kira.h"
#include "tools.h"
#include "dataBase.h"
#include "createSeeds.h"
#include <omp.h>
#include <pthread.h>
#include <sstream> 
#include <tuple>

using namespace std;
using namespace GiNaC;

static Loginfo & logger = Loginfo::instance();

static pthread_mutex_t mPreserveFermat;
static pthread_cond_t cPreserveFermat;

static pthread_mutex_t mPreserveResults;
static pthread_cond_t cPreserveResults;

void Kira::initiate_fermat(int kira2Math=0) {

  char var[40], token[40];

  var[0]='\0';
  sprintf(var,"d\n");

  for(size_t i=0;i<invar.size();i++){
    if ( something_string(invar[i])==something_string(massSet2One) 
	 && !kira2Math )
      continue;
    sprintf(token,"%s\n",something_string(invar[i]).c_str());
    strcat(var,token);
  }
  core_mask = ((1LL<<coreNumber)-1LL);
  
  fermat = new Fermat[coreNumber];
  for(int i=0; i<coreNumber; i++){
    fermat[i].start_fermat(fermatPath,var);
  }

  pthread_mutex_init(&mPreserveFermat, NULL);
  pthread_cond_init (&cPreserveFermat, NULL);
  pthread_mutex_init(&mPreserveResults, NULL);
  pthread_cond_init (&cPreserveResults, NULL);
  
}

void Kira::destroy_fermat() {
  
  pthread_mutex_destroy(&mPreserveFermat);
  pthread_cond_destroy(&cPreserveFermat);
  pthread_mutex_destroy(&mPreserveResults);
  pthread_cond_destroy (&cPreserveResults);
  
  delete [] fermat;
  
}


void fermat2flushTimes(string & a, string & b, string & c, Fermat &fermat) {
  
  fermat.fermat_collect(const_cast<char*>("-("));
  fermat.fermat_collect(const_cast<char*>(a.c_str()));
  fermat.fermat_collect(const_cast<char*>(")/("));
  fermat.fermat_collect(const_cast<char*>(b.c_str()));
  fermat.fermat_collect(const_cast<char*>(")*("));
  fermat.fermat_collect(const_cast<char*>(c.c_str()));
  fermat.fermat_collect(const_cast<char*>(")"));
  
};

void fermat2flushPlus(string & d,Fermat &fermat) {
  
  fermat.fermat_collect(const_cast<char*>("+("));
  fermat.fermat_collect(const_cast<char*>(d.c_str()));
  fermat.fermat_collect(const_cast<char*>(")"));
  
};

bool myfunction (const string &i,const string &j) { return (i.length()<j.length()); };

int Kira::calculate_coefficient_term(list<string>& ptrToString) {
  
  omp_lock_t lck;
  omp_init_lock(&lck);
  
  int numberOfCoefficients=ptrToString.size();
  string mth;
  int id=-1;
  if(numberOfCoefficients==1){
    pthread_mutex_lock(&mPreserveFermat);
    while(!core_mask)
      pthread_cond_wait(&cPreserveFermat, &mPreserveFermat);

    for(int i=0;i<coreNumber;i++){
      if(((core_mask>>i)&1LL)==1LL){
	id=i;
	core_mask &= ~(1LL<<i);
	break;
      }
    }
    pthread_mutex_unlock(&mPreserveFermat);

    if(id>=0){
      fermat[id].fermat_collect(const_cast<char*>((ptrToString.front()).c_str()));
      ptrToString.pop_front();
      fermat[id].fermat_calc();
      mth.assign(fermat[id].g_baseout);
      ptrToString.push_front(mth);
      pthread_mutex_lock(&mPreserveFermat);
	  core_mask |= 1LL<<id;
      pthread_cond_broadcast(&cPreserveFermat);
      pthread_mutex_unlock(&mPreserveFermat);
      return 1;
    }
  }
  
  omp_set_dynamic(0);
  int nThreads = ((numberOfCoefficients/3)<coreNumber)? (numberOfCoefficients/3) : coreNumber;
  omp_set_num_threads(nThreads);

  #pragma omp parallel shared(lck,ptrToString,numberOfCoefficients) private(id,mth)
  {
    id=-1;   
    while(numberOfCoefficients>1){
      omp_set_lock(&lck);
        numberOfCoefficients=(ptrToString.size());
        id=-1;
        if(numberOfCoefficients>1){
          pthread_mutex_lock(&mPreserveFermat);
          while(!core_mask){
            pthread_cond_wait(&cPreserveFermat, &mPreserveFermat);
          }

          for(int i=0;i<coreNumber;i++){
            if(((core_mask>>i)&1LL)==1LL){
              id=i;
              core_mask &= ~(1LL<<i);
              break;
            }
          }
          pthread_mutex_unlock(&mPreserveFermat);
          if(id>=0){
            fermat[id].fermat_collect(const_cast<char*>((ptrToString.front()).c_str()));
            ptrToString.pop_front();
            fermat[id].fermat_collect(const_cast<char*>("+"));
            fermat[id].fermat_collect(const_cast<char*>((ptrToString.front()).c_str()));
            ptrToString.pop_front();
          }
        }
      omp_unset_lock(&lck);
      
      if(id>=0){
        
        fermat[id].fermat_calc();
        
        mth.assign(fermat[id].g_baseout);
        
        pthread_mutex_lock(&mPreserveFermat);
          core_mask |= 1LL<<id;
        pthread_cond_broadcast(&cPreserveFermat);
        pthread_mutex_unlock(&mPreserveFermat);
        
        omp_set_lock(&lck);
          ptrToString.push_front(mth);
          ptrToString.sort(myfunction);
          numberOfCoefficients=(ptrToString.size());
        omp_unset_lock(&lck);
      }
      id=-1;
    }
  }
  omp_destroy_lock(&lck);

  return 1;
  
};

int Kira::calculate_coefficient(list<string>& ptrToString) {
  
  int id=-1;
  pthread_mutex_lock(&mPreserveFermat);
  while(!core_mask)
    pthread_cond_wait(&cPreserveFermat, &mPreserveFermat);
  for(int i=0;i<coreNumber;i++){
    if(((core_mask>>i)&1LL)==1LL){
      id=i;
      core_mask &= ~(1LL<<i);
      break;
    }
  }
  pthread_mutex_unlock(&mPreserveFermat);
  if(id>=0){
    for(list<string>::iterator it = ptrToString.begin();
      it !=ptrToString.end();++it){
      if (it != ptrToString.begin()){
	fermat[id].fermat_collect(const_cast<char*>("+"));
      }
      fermat[id].fermat_collect(const_cast<char*>((*it).c_str()));
    }
  }
  return id;
  
};

int Kira::insert_equation(Integral *&EqPtr, Integral *&insertEqPtr, int id) {
  
  int srange = EqPtr[0].length;
  int rrange = insertEqPtr[0].length;
  int kilo = srange;
  
  // maximum length of the new equation
  srange = rrange - 2 + srange;
  Integral* newEqPtr = new Integral[srange+1];
  int* merke = new int[srange+1];
  for (int i = 0; i < kilo; i++) {
    newEqPtr[i]= EqPtr[i];
    merke[i]=0;
  }
  delete [] EqPtr;

  swap(newEqPtr[0], newEqPtr[kilo-1]);
  int j = kilo-1;
  newEqPtr[srange].id = -1;
  newEqPtr[srange].coefficientString.clear();
  swap( newEqPtr[srange], newEqPtr[kilo-1] );
  int csrange = srange;
  for (int k = 1; k < rrange; k++) {
    merke[j]=1;
    newEqPtr[j] = insertEqPtr[k];
    j++;
  }
  // coefficients of the same integrals will be collected
  for (int itt = 0; itt < srange-1; itt++){
    for (int ittt = itt + 1; ittt < srange; ittt++){
      if (newEqPtr[itt].id == newEqPtr[ittt].id){
        swap(newEqPtr[ittt], newEqPtr[srange-1]);
        swap(merke[ittt], merke[srange-1]);
        if(merke[itt]){
	  fermat2flushTimes(newEqPtr[itt].coefficientString,insertEqPtr[0].coefficientString,
	    newEqPtr[csrange].coefficientString,fermat[id]);
	  fermat2flushPlus(newEqPtr[srange-1].coefficientString,fermat[id]);
	  fermat[id].fermat_calc();
          newEqPtr[itt].coefficientString.assign(fermat[id].g_baseout);
          merke[itt]=0;
        }
        else{
	  fermat2flushTimes(newEqPtr[srange-1].coefficientString,insertEqPtr[0].coefficientString,
	   newEqPtr[csrange].coefficientString,fermat[id]);
	  fermat2flushPlus(newEqPtr[itt].coefficientString,fermat[id]);
          fermat[id].fermat_calc();
          newEqPtr[itt].coefficientString.assign(fermat[id].g_baseout);
          merke[srange-1]=0;
        }
        newEqPtr[srange-1].coefficientString.clear();
        --srange;
        if (newEqPtr[itt].coefficientString == "0"){
          swap( newEqPtr[itt], newEqPtr[srange-1] );
          swap( merke[itt], merke[srange-1]);
          newEqPtr[srange-1].coefficientString.clear();
          --srange;
          --itt;
        }
        break;
      }
    }
  }

  for (int itt = 0; itt < srange; itt++){
    if(merke[itt]){
      fermat2flushTimes(newEqPtr[itt].coefficientString,insertEqPtr[0].coefficientString,
      newEqPtr[csrange].coefficientString,fermat[id]);
      fermat[id].fermat_calc();
      newEqPtr[itt].coefficientString.assign(fermat[id].g_baseout);
      merke[itt]=0;
    }
  }

  if (srange != 0){
    newEqPtr[0].length=srange;
    EqPtr = new Integral[srange];
    for (int i = 0; i < srange; i++)
      EqPtr[i] = newEqPtr[i];
  }
  
  delete [] newEqPtr;
  newEqPtr=0;
  delete [] merke;
  
  if (srange != 0)
    return 1;
  else
    return 0;
}

void Kira::sort_equation(Integral *&EqPtr) {

  int srange=EqPtr[0].length;

  for (int i = 0; i < srange-1; i++) {
    for (int ii = i+1; ii < srange; ii++) {
      if ( EqPtr[ii].id > EqPtr[i].id ) {
        swap(EqPtr[i], EqPtr[ii]);
      }
    }
  }
  
  EqPtr[0].length=srange;
  
#ifdef KIRA_DEBUG
  if (srange == 0){
    logger.set_level(2);
    logger<<"This equation is zero: \n";
    logger.set_level(1);
  }
  else if (srange < 2) {
    logger.set_level(2);
    logger<<"This integral is zero: \n";
    for (int i = 0; i < srange; i++) {
      logger<<"Int[" << EqPtr[i].id << "," << EqPtr[i].flag2 <<"]" <<"*";
      logger<< EqPtr[i].coefficientString << "\n";
    }
    logger<<"--------------------------------" << "\n";
    logger.set_level(1);
  }
#endif
}

void Kira::run_reduction(int opt) {

  ofstream myfile;
  double finalt = 0;

  int information = 0, red_1 = -1, whileloop = 0;

  logger << "\nIterate the system until it is in triangular form: \n";

  int counter = 0;  
  do{
    Clock clock;
    
    red_1 = 0;
    
    logger << "iteration: " << whileloop << "; ";
    logger << "items left: " << reduce_please.size() << "; ";
    
    omp_set_dynamic(0);
    int nThreads = coreNumber;
    omp_set_num_threads(nThreads);

    #pragma omp parallel for schedule(dynamic)
    for( vector<long unsigned>::iterator it = reduce_please.begin(); it < reduce_please.end(); it++){
      
      Integral *** fwEQ;
      unsigned lengthEQ;
    
      int id = omp_get_thread_num();
      
      pthread_mutex_lock(&mPreserveResults);
        while( ((rdy2F[*it]>>0)&1) )
          pthread_cond_wait(&cPreserveResults, &mPreserveResults);
        rdy2F[(*it)] |= 1<<0;
          
        FWREDI mapContent = forwardRed.find(*it);
        std::tie(fwEQ,lengthEQ) = mapContent->second;
        get<1>(mapContent->second)=1;
        
      pthread_mutex_unlock(&mPreserveResults);
      
      for (unsigned k=1; k < lengthEQ; k++){
        
        int ende54 = 0;
        int grenze = ( fwEQ[0][0][0].length<fwEQ[0][k][0].length ) ?
                      fwEQ[0][0][0].length : fwEQ[0][k][0].length;
        
        if ( ende54 == 0 && fwEQ[0][0][0].length >
          fwEQ[0][k][0].length){
          swap(fwEQ[0][0],fwEQ[0][k]);
          ende54 = 1;
        }
        if ( ende54 == 0 && fwEQ[0][0][0].length <
          fwEQ[0][k][0].length){
          ende54 = 1;
        }
        
        if(ende54 != 0)
          continue;
        
        for (int i = 1; i < grenze; i++) {
          if ( (fwEQ[0][0][i].id > fwEQ[0][k][i].id ) ) {
            swap(fwEQ[0][0],fwEQ[0][k]);
            ende54 = 1;
            break;
          }
          if ( (fwEQ[0][0][i].id <fwEQ[0][k][i].id) ) {
            ende54 = 1;
            break;
          }
        }
        
      }
      
      for(int jtt=lengthEQ-1;jtt > 0;jtt--){
        int raw = 0;
        if( (raw = insert_equation(fwEQ[0][jtt], fwEQ[0][0],id)) ){
          
          sort_equation(fwEQ[0][jtt]);
          long unsigned ID = fwEQ[0][jtt][0].id;
          pthread_mutex_lock(&mPreserveResults);
            
            RDY2FI rdy2FContent = rdy2F.find(ID);
            
            if( !(rdy2FContent!=rdy2F.end()) ){
              rdy2F.insert(pair<long unsigned,int> (ID,0));
            }
          
            while( ((rdy2F[ID]>>0)&1) )
              pthread_cond_wait(&cPreserveResults, &mPreserveResults);
            rdy2F[ID] |= 1<<0;
            
            FWREDI mapContent = forwardRed.find(ID);
            if ( mapContent != forwardRed.end() ) {
              
              get<0>(mapContent->second)[0][get<1>(mapContent->second)]=fwEQ[0][jtt];
              get<1>(mapContent->second)++;
              red_1++;
            }
            else{
              Integral *** put2map = new Integral ** [1];
              put2map[0] = new Integral * [RED];
              put2map[0][0]=fwEQ[0][jtt];
              forwardRed.insert(pair<long unsigned, tuple<Integral***, int> > (ID, make_tuple(put2map,1)));
            }
          pthread_mutex_unlock(&mPreserveResults);
          
          #pragma omp critical
          {
            information++;
          }
          pthread_mutex_lock(&mPreserveResults);
            rdy2F[ID] &= ~(1<<0);
          pthread_cond_broadcast(&cPreserveResults);
          pthread_mutex_unlock(&mPreserveResults);
        }
        else{
          #pragma omp critical
          {
            counter++;
          }
        }
        lengthEQ--;
        
      }
      pthread_mutex_lock(&mPreserveResults);
        rdy2F[(*it)] &= ~(1<<0);
      pthread_cond_broadcast(&cPreserveResults);
      pthread_mutex_unlock(&mPreserveResults);
    }

    reduce_please.clear();
    for( FWREDI toIt = forwardRed.begin(); toIt != forwardRed.end(); toIt++){ //collect all keys to iterate
      if(get<1>(toIt->second)>1){
        reduce_please.push_back(get<0>(toIt->second)[0][0]->id);
      }
    }

    if(red_1 == 0){
      for( FWREDI toIt = forwardRed.begin(); toIt != forwardRed.end(); toIt++){ //collect all keys to iterate
        if(get<1>(toIt->second)>0){
          reduce_please.push_back(get<0>(toIt->second)[0][0]->id);
        }
      }
    }
    
    finalt+=clock.eval_time();
    
    whileloop++;
    
    logger << "time: ( " << finalt << " s )\n";
    
  } while(red_1);
  
  logger << "redundant: " << counter<<"; insertions "<<information<<"\n";
  // Triangular form is complete, norm the coefficients and write them to the
  // hard disk.
  Clock clock;
  
  int countbad = 0, countgood = 0;
  logger << "All coefficients will be normalized\n";
  
  omp_set_dynamic(0);
  int nThreads = coreNumber;
  omp_set_num_threads(nThreads);
  
  #pragma omp parallel for schedule(dynamic)
  for( vector<long unsigned>::iterator it = reduce_please.begin(); it < reduce_please.end(); it++){
    Integral*** fwEQ;
    unsigned lengthEq;
    FWREDI mapContent = forwardRed.find(*it);
    tie(fwEQ,lengthEq) = mapContent->second;
    
    if(lengthEq && fwEQ[0][0][0].flag2 == 0){
      
      int id = omp_get_thread_num();
      
      #pragma omp critical
      {
        countgood++;
      }
      for (unsigned j = 0; j< fwEQ[0][0][0].length; j++) {
        if(j!=0){
          fermat[id].fermat_collect(const_cast<char*>("-("));
          fermat[id].fermat_collect(const_cast<char*>(fwEQ[0][0][j].coefficientString.c_str()));
          fermat[id].fermat_collect(const_cast<char*>(")/("));
          fermat[id].fermat_collect(const_cast<char*>(fwEQ[0][0][0].coefficientString.c_str()));
          fermat[id].fermat_collect(const_cast<char*>(")"));
          fermat[id].fermat_calc();
          fwEQ[0][0][j].coefficientString.assign(fermat[id].g_baseout);
        }
      }
      fwEQ[0][0][0].coefficientString = "1";
    }
    else if(lengthEq && fwEQ[0][0][0].flag2 == -1){
      #pragma omp critical
      {
        countbad++;
      }
      for (unsigned j = 0; j<fwEQ[0][0][0].length; j++) {
        fwEQ[0][0][j].coefficientString.clear();
      }
      delete [] fwEQ[0][0];
      fwEQ[0][0] = 0;
      delete [] fwEQ[0];
      fwEQ[0] = 0;
      get<1>(mapContent->second)=0;
    }
  }
    
  finalt+=clock.eval_time();
  logger << "\nEquations out of bounds rmax and smax: " << countbad << "\n";
  logger << "Equations in the bounds rmax and smax: " << countgood << "\n";
  logger << "( " << finalt << " s )" << "\n";
  
}

void write_result(Integral *&EqPtr,int i_Integrals,std::ofstream &output) {
 
  output<<"- Eq: "<<std::endl;

  for(int i = 0; i < i_Integrals; i++) {
    output << "  - ["<<EqPtr[i].id<<",";
    output << EqPtr[i].flag2 << ",";
    output << EqPtr[i].characteristics[SECTOR] << ",";
    output << EqPtr[i].characteristics[TOPOLOGY] << ",";
    output << EqPtr[0].length << ",";
    output << "\"";
    output<<EqPtr[i].coefficientString;
    output<<"\"]" << std::endl;
  }
};

int Kira::get_number_of_masterintegrals(unsigned it,vector<long unsigned>& finalResults, 
					vector<long unsigned>& lriMI, int **IDarray) {
  
  vector<int> subsInt;
  int lritID=0;
  int countReducedIntegral = 0;
  
  long unsigned lrit = last_reduce[it];
  
  pthread_mutex_lock(&mPreserveResults);
    auto ocContent = occurrence.find(lrit);
    if(ocContent != occurrence.end()){
      (ocContent->second)--;
      if(ocContent->second==0){
        if(serialWrite == 0)
          finalResults.push_back(lrit);
        countReducedIntegral++;
      }
    }
    for(unsigned ktt=1; ktt < allEq[it][0][0].length; ktt++){
      long unsigned ID = allEq[it][0][ktt].id;
      auto itID = reverseLastReduce.find(ID);
      if(itID != reverseLastReduce.end()){
        unsigned eqnum = itID->second;
        while( ((rdy2P[eqnum]>>0)&1) )
	  pthread_cond_wait(&cPreserveResults, &mPreserveResults);
      }
    }
  pthread_mutex_unlock(&mPreserveResults);
  
  
  for(unsigned ktt=1; ktt < allEq[it][0][0].length; ktt++){
    long unsigned ID = allEq[it][0][ktt].id;
    auto itID = reverseLastReduce.find(ID);
    if(itID != reverseLastReduce.end()){
      unsigned eqnum = itID->second;
      lritID++;//count non master integrals
      IDarray[ktt]= new int [allEq[eqnum][0][0].length];
      IDarray[ktt][0]=allEq[eqnum][0][0].length;
      for(int i=1;i<IDarray[ktt][0];i++){
        IDarray[ktt][i]=i;
        lriMI.push_back(allEq[eqnum][0][i].id);
      }
    }
    else{
      IDarray[ktt]= new int [1];
      IDarray[ktt][0]=0;
      subsInt.push_back(ktt);// book keeping of masterintegrals
      lriMI.push_back(ID);
    }
  }
  
  lritID=allEq[it][0][0].length-lritID;//Number of masterintegrals in the first equation
  IDarray[0]= new int [lritID];
  IDarray[0][0]=lritID;
  for(int i=1;i<IDarray[0][0];i++){
    IDarray[0][i]=subsInt[i-1];
  }
  sort(lriMI.begin(),lriMI.end());
  lriMI.resize(distance(lriMI.begin(),
			unique(lriMI.begin(), lriMI.end())) );
  
  return countReducedIntegral;
  
}

unsigned Kira::addup_common_masters(unsigned it, unsigned j, int **IDarray, 
				Integral newEqPtr[], vector<list<string> >  &listPtrToString) {
      
  // Add up all master integrals, which exist in [it] first.
  if( IDarray[0][0]>1){
    for(int jt=IDarray[0][0]-1; jt>0; jt--){
      int JD=IDarray[0][jt];
      
      list<string>  ptrToString;
      string tokenStr="(";
      tokenStr+= allEq[it][0][JD].coefficientString+")";
      ptrToString.push_back(tokenStr);
      
      swap(IDarray[0][jt],IDarray[0][ (IDarray[0][0]-1) ]);
      IDarray[0][0]--;
      
//Check the id of the masterintegral[JD] in eq[it] is the same as in eq[ID2].
//If yes then add up this masterintegral which exists in all eq[ID2]
      for(unsigned it2=1; it2<allEq[it][0][0].length; it2++){
	if(IDarray[it2][0]){
	  long unsigned ID2 = allEq[it][0][it2].id;
          auto itID = reverseLastReduce.find(ID2);
            unsigned eqnum = itID->second;
	  //eq which will be substituted: (master integrals)
	  for(int jt2=IDarray[it2][0]-1; jt2>0; jt2--){
	    int JD2=IDarray[it2][jt2];
	    if(allEq[eqnum][0][JD2].id == allEq[it][0][JD].id){
	      tokenStr="(";
	      tokenStr+= allEq[eqnum][0][JD2].coefficientString+")*("
	      + allEq[it][0][it2].coefficientString+")";
	      ptrToString.push_back(tokenStr);
	      swap(IDarray[it2][jt2],IDarray[it2][ (IDarray[it2][0]-1) ]);
	      IDarray[it2][0]--;
	    }
	  }
	}
      }
      listPtrToString.push_back(ptrToString);
      newEqPtr[j]=allEq[it][0][JD];
      j++;
    }
  }
  return j;
}

unsigned Kira::addup_final_masters(unsigned it, unsigned j, int **IDarray, 
				Integral newEqPtr[], vector<list<string> >  &listPtrToString) {
  
  //Add up coefficients of the integrals with the same id: JD
  for(unsigned itK=1; itK<allEq[it][0][0].length; itK++){
    if( IDarray[itK][0]>1){
      long unsigned ID = allEq[it][0][itK].id;
      auto itID = reverseLastReduce.find(ID);
        unsigned eqnum = itID->second;
      
      for(int jt=IDarray[itK][0]-1; jt>0; jt--){
	int JD=IDarray[itK][jt];
	
	list<string>  ptrToString;
	string tokenStr="(";
	tokenStr+= allEq[eqnum][0][JD].coefficientString+")*("
	+ allEq[it][0][itK].coefficientString+")";
	ptrToString.push_back(tokenStr);

	swap(IDarray[itK][jt],IDarray[itK][ (IDarray[itK][0]-1) ]);
	IDarray[itK][0]--;
	for(unsigned it2=itK+1; it2<allEq[it][0][0].length; it2++){
	  if( IDarray[it2][0]){
                     
	    long unsigned ID2 = allEq[it][0][it2].id;
            auto itID2 = reverseLastReduce.find(ID2);
              unsigned eqnum2 = itID2->second;
	    for(int jt2=IDarray[it2][0]-1; jt2>0; jt2--){
	      int JD2=IDarray[it2][jt2];
	      if(allEq[eqnum2][0][JD2].id == allEq[eqnum][0][JD].id){
		
		tokenStr="(";
		tokenStr+= allEq[eqnum2][0][JD2].coefficientString+")*("
		+ allEq[it][0][it2].coefficientString+")";
		ptrToString.push_back(tokenStr);
		
		swap(IDarray[it2][jt2],IDarray[it2][ (IDarray[it2][0]-1) ]);
		IDarray[it2][0]--;
	      }
	    }
	  }
	}
	listPtrToString.push_back(ptrToString);
	newEqPtr[j]=allEq[eqnum][0][JD];
	j++;
      }
    }
  }
  return j;
}

void Kira::back_subs() {
  
  Clock clock;
  logger <<"\nBack substitution: \n";
  ofstream myfile;  
  
  int maxNThreads=static_cast<int>(reduct2StartHere.size());
  omp_set_nested(1);
  int initNThreads = (maxNThreads<coreNumber)? maxNThreads : coreNumber;
  omp_set_num_threads(initNThreads);
  int allIterations=0;
  for(int it=0;it<initNThreads;it++){
    allIterations += (eqnum-reduct2StartHere[it]);
  }
  int iterationCount=0;
  int reducedIntegralCount=0;
  if(dataFile == 1){
    string nameResult = outputDir+"/results/"+integralfamily.name+"/kira";
    myfile.open (nameResult.c_str());
  }
  std::set<long unsigned> masterIntegral;
  
  DataBase database(outputDir+"/results/"+integralfamily.name+"/kira.db");
  
  
  if(dataBase == 1)
    database.create_equation_table();
  
  double lastTime = 600;
  
#pragma omp parallel shared(iterationCount,reducedIntegralCount,myfile,masterIntegral,lastTime)
  {
    vector<long unsigned> finalResults;
    unsigned idThread = omp_get_thread_num();
    unsigned itStart = reduct2StartHere[idThread];
    
    for( unsigned it = itStart;  it < reihen; it++){

      int pass2R = 0;
      
      pthread_mutex_lock(&mPreserveResults);
        if( ((rdy2P[it]>>1)&1) ){
          pass2R = 1;
          rdy2P[it] &= ~(1<<1);
        }
      pthread_mutex_unlock(&mPreserveResults);
      
      if(!pass2R){
	#pragma omp critical
	{
	  load_bar(++iterationCount,allIterations,1,1000);
	}
	continue;
      }
      
      vector<long unsigned> lriMI; 
      int **IDarray= new int* [allEq[it][0][0].length];
      
      int token = 
      get_number_of_masterintegrals(it,finalResults,lriMI,IDarray);
      
      pthread_mutex_lock(&mPreserveResults);
	reducedIntegralCount+=token;
      pthread_mutex_unlock(&mPreserveResults);
      
      Integral* newEqPtr = new Integral[lriMI.size()+1];
      unsigned j=0;
      newEqPtr[j]=allEq[it][0][0];
      j++;
      
      vector<list<string> > listPtrToString;
      
      j = addup_common_masters(it,j,IDarray,newEqPtr,listPtrToString);
      j = addup_final_masters(it,j,IDarray,newEqPtr,listPtrToString);
      
      pthread_mutex_lock(&mPreserveResults);
	for(unsigned ktt=1; ktt < allEq[it][0][0].length; ktt++){
	  long unsigned ID = allEq[it][0][ktt].id;
          auto ocContent = occurrence.find(ID);
          if(ocContent != occurrence.end())
            (ocContent->second)--;
          auto itID = reverseLastReduce.find(ID);
          if((ocContent->second) == 0 && itID != reverseLastReduce.end()){
            
            if(serialWrite == 0)
              finalResults.push_back(ID);
	    reducedIntegralCount++;
	  }
	}
      pthread_mutex_unlock(&mPreserveResults);
            
      int numberOfCoefficients=(listPtrToString.size());
      int nThreads = (numberOfCoefficients<coreNumber)? numberOfCoefficients : coreNumber;
      omp_set_num_threads(nThreads);
      
      #pragma omp parallel for schedule(dynamic)
      for(int jIT=(numberOfCoefficients-1); jIT>-1;jIT--) {
        
        listPtrToString[jIT].sort(myfunction);
        
        if(listPtrToString[jIT].size()>5 && algebra){
          
	  calculate_coefficient_term(listPtrToString[jIT]);
	  newEqPtr[jIT+1].coefficientString=listPtrToString[jIT].front();
	}
	else{
          
	  int id = -1;
          
	  id=calculate_coefficient(listPtrToString[jIT]);
	  
          if(id<0) {
	    logger <<"Fermat has problems.";
	    exit(-1);
	  }
	  fermat[id].fermat_calc();
	  newEqPtr[jIT+1].coefficientString.assign(fermat[id].g_baseout);
	  pthread_mutex_lock(&mPreserveFermat);
	    core_mask |= 1LL<<id;
	  pthread_cond_broadcast(&cPreserveFermat);
	  pthread_mutex_unlock(&mPreserveFermat);
	}
      }
      
      for(unsigned jIT=1; jIT<j; jIT++){
	if(newEqPtr[jIT].coefficientString=="0"){
	  swap(newEqPtr[jIT],newEqPtr[j-1]);
	  j--;
	  jIT--;
	}
      }
      
      
      pthread_mutex_lock(&mPreserveResults);
	for(unsigned ktt=0; ktt < allEq[it][0][0].length; ktt++)
	  delete [] IDarray[ktt];
	delete [] IDarray;
	newEqPtr[0].length = j;
	swap(newEqPtr,allEq[it][0]);
	delete [] newEqPtr;
	newEqPtr=0;
	sort_equation(allEq[it][0]);
	rdy2P[it] &= ~(1<<0);
	
        double nowTime=clock.eval_time();
        
        if(lastTime<(nowTime-600)){
	  
          lastTime = nowTime;
        
          if(finalResults.size()>0 && serialWrite == 0){
            
            if(dataBase == 1){
              database.begin_transaction();
              database.prepare_backsubstitution();
            }
            
            for ( size_t itfR = 0, end1 = finalResults.size(); itfR < end1; itfR++) {
              long unsigned ID = finalResults[itfR];
              auto itID = reverseLastReduce.find(ID);
              if(itID != reverseLastReduce.end()){
                unsigned eqnum = itID->second;
                for(unsigned i = 0; i < allEq[eqnum][0][0].length; i++) {
                  if (i>0) {
                    masterIntegral.insert(allEq[eqnum][0][i].id);
                  }
                  if(dataBase == 1)
                    database.bind_equation(allEq[eqnum][0][i], allEq[eqnum][0][0].length, ID);
                }
                if(dataFile == 1)
                  write_result(allEq[eqnum][0], allEq[eqnum][0][0].length, myfile);
                
                delete [] allEq[eqnum][0];
                allEq[eqnum][0]=0;
                delete [] allEq[eqnum];
                allEq[eqnum]=0;
                length[eqnum]=0;
              }
            }
            
            finalResults.clear();
            
            if(dataBase == 1)
              database.commit_transaction();
	    
            logger << "\n\nKira committed results\n";
	  }
	}
      pthread_cond_broadcast(&cPreserveResults);
      pthread_mutex_unlock(&mPreserveResults);
#pragma omp critical
      {
	load_bar(++iterationCount,allIterations,1,1000);
      }
    }
    
    
    pthread_mutex_lock(&mPreserveResults);
      if(finalResults.size() > 0 && serialWrite == 0){
          
        if(dataBase == 1){
          database.begin_transaction();
          database.prepare_backsubstitution();
        }
        for ( size_t itfR = 0, end1 = finalResults.size(); itfR != end1; itfR++) {
          long unsigned ID = finalResults[itfR];
          auto itID = reverseLastReduce.find(ID);
          if( itID != reverseLastReduce.end()){
            unsigned eqnum = itID->second;
            for(unsigned i = 0; i < allEq[eqnum][0][0].length; i++) {
              if (i>0) {
                masterIntegral.insert(allEq[eqnum][0][i].id);
              }
              if(dataBase == 1)
                database.bind_equation(allEq[eqnum][0][i], allEq[eqnum][0][0].length, ID);
            }
            if(dataFile == 1)
              write_result(allEq[eqnum][0], allEq[eqnum][0][0].length, myfile);
            
            delete [] allEq[eqnum][0];
            allEq[eqnum][0]=0;
            delete [] allEq[eqnum];
            allEq[eqnum]=0;
            length[eqnum]=0;
          }
        }
        
        finalResults.clear();
        
        if(dataBase == 1)
          database.commit_transaction();
        
        logger << "\n\nKira committed results\n\n";
      }
    pthread_cond_broadcast(&cPreserveResults);
    pthread_mutex_unlock(&mPreserveResults);
  }
  
// #### Reduction is complete. Write it on hard disk (serialWrite). ####
  
  if(serialWrite == 1){
    
    logger <<"\nWrite the reduction results in "<<outputDir<<"/results/ (serial)\n";
    
    if(dataBase == 1){
      database.begin_transaction();
      database.prepare_backsubstitution();
    }
    
    for ( unsigned it = 0, end1 = reihen; it < end1; it++) {
      long unsigned lrit = last_reduce[it];
      if(length[it]){
        for(unsigned i = 0; i < allEq[it][0][0].length; i++) {
          if (i>0) {
            masterIntegral.insert(allEq[it][0][i].id);
          }
          if(dataBase == 1)
            database.bind_equation(allEq[it][0][i], allEq[it][0][0].length, lrit);
        }
        
        if(dataFile == 1){
          write_result(allEq[it][0], allEq[it][0][0].length, myfile);
        }
      }
      load_bar(it+1,eqnum,50,100);
    }
    if( dataBase == 1 )
      database.commit_transaction();
  }
  

  if(dataFile == 1)
    myfile.close();
  
  string pathMI = outputDir+"/results/"+integralfamily.name+"/masters.final";
  myfile.open (pathMI.c_str());
  logger <<"\nNumber of master integrals: " << masterIntegral.size()<<"\n";
   
  for ( auto itg = masterIntegral.begin(); itg != masterIntegral.end(); itg++ ){
    tuple<string,unsigned,unsigned> integral;
    get_properties(*(itg),integral);
    logger << *(itg) << ": ";
    logger << collectReductions[get<1>(integral)] << ": ";
    logger << get<0>(integral) << ": ";
    logger << get<2>(integral);
    logger << "\n";
    
    myfile << *(itg) << ": ";
    myfile << collectReductions[get<1>(integral)] << ": ";
    myfile << get<0>(integral) << ": ";
    myfile << get<2>(integral);
    myfile << "\n";
  }
  
  logger << reducedIntegralCount<<" equations were reduced\n";
  logger << "( " << clock.eval_time() << " s )\n";
  
}